//****************************************************************************************//
//Komponenty do testowania, wysyania i odbioru UPO JPK - Jednolitego Pliku Kontrolnego **//
//autor: IPS Przedsibiorstwo Informatyczne - Marcin Konopka 2016r. **********************//
//****************************************************************************************//

unit JPKXMLCompUnit;

interface

uses
  Windows, SysUtils, Registry, Classes;

{$DEFINE NOTypesDef}
type
  {$IFDEF UNICODE}
  MainString = String;
  MainAnsiString = AnsiString;
  PMainAnsiChar = PAnsiChar;
  PMainChar = PChar;
  {$ELSE}
  MainString = WideString;
  MainAnsiString = String;
  PMainAnsiChar = PChar;
  PMainChar = PWideChar;
  {$ENDIF}
  TDocumentType = (dt_JPK,dt_JPKAH);

  TError = class
    err:Integer;
    errMsg:String;
  end;

  TJPKXMLNoXMLJPKException = class(Exception)

  end;

  TCustomJPKXML = class(TComponent)
  private
    FReferenceNumber: MainString;
    FStatus: Integer;
    FStatusText: String;
    FUPO: String;
    FDocumentType: TDocumentType;
    FErr: TError;
    FTestMode: Boolean;
    FLockRefreshUPO: Boolean;
    procedure SetReferenceNumber(const Value: MainString);
    procedure SetStatus(const Value: Integer);
    procedure SetUPO(const Value: String);
    procedure SetDocumentType(const Value: TDocumentType);
    procedure SetStatusText(const Value: String);
    function GetError: TError;
    function GetLicText: String;
    function GetParentWinHandle: HWND;
    procedure SetLicText(const Value: String);
    procedure SetParentWinHandle(const Value: HWND);
    procedure SetTestMode(const Value: Boolean);
  protected
    function GetFileName_:MainString; virtual; abstract;
    function GetData_:MainAnsiString; virtual; abstract;
    function IsIntData: Boolean; virtual; abstract;
    procedure AddJPKError;
  public
    destructor Destroy; override;
    property LicText:String read GetLicText write SetLicText;
    property ParentWinHandle:HWND read GetParentWinHandle write SetParentWinHandle;
    property ReferenceNumber:MainString read FReferenceNumber write SetReferenceNumber;
    property Status:Integer read FStatus write SetStatus;
    property UPO:String read FUPO write SetUPO;
    property StatusText:String read FStatusText write SetStatusText;
    property DocumentType:TDocumentType read FDocumentType write SetDocumentType;
    property TestMode:Boolean read FTestMode write SetTestMode;
    procedure RefreshUPO(Break:Boolean=false);
    property Error:TError read GetError;
    procedure AddError(AErr:Integer; AErrMsg:String);
  end;

  TJPKXMLFile = class(TCustomJPKXML)
  private
    FFileNamePath: MainString;
    procedure SetFileNamePath(const Value: MainString);
    { Private declarations }
  protected
    { Protected declarations }
    function GetFileName_:MainString; override;
    function GetData_:MainAnsiString; override;
    function IsIntData: Boolean; override;
  public
    { Public declarations }
  published
    { Published declarations }
    property LicText;
    property FileNamePath:MainString read FFileNamePath write SetFileNamePath;
    property ReferenceNumber;
    property Status;
    property UPO;
    property TestMode;
  end;

  TJPKXMLList = class
  private
    FList: TStringList;
    FOwnsObj: Boolean;
    function GetItems(index: Integer): TCustomJPKXML;
  public
    constructor Create(AOwnsObj:Boolean);
    destructor Destroy; override;
    procedure Add(JPKXML:TCustomJPKXML);
    procedure Clear;
    function Count:Integer;
    property Items[index:Integer]:TCustomJPKXML read GetItems; default;
    procedure Delete(index:Integer);
  end;

  TCustomJPKXMLSender = class;

  TJPKXMLFinishSend = procedure(Sender:TCustomJPKXMLSender; JPKXML:TCustomJPKXML) of object;

  TCustomJPKXMLSender = class(TComponent)
  private
    FJPKXMLList: TJPKXMLList;
    FOnFinishSend: TJPKXMLFinishSend;
    FSessionID: Pointer;
    FTestMode: Boolean;
    FSyncOnFinisch: Boolean;
    FIsUserBreak: Boolean;
    FFinishIndex: Integer;
    procedure SetSessionID(const Value: Pointer);
    procedure SetTestMode(const Value: Boolean);
    procedure SetSyncOnFinisch(const Value: Boolean);
    function GetIsUserBreak: Boolean;
    function GetLicText: String;
    function GetParentWinHandle: HWND;
    procedure SetLicText(const Value: String);
    procedure SetParentWinHandle(const Value: HWND);
  protected
    property SessionID:Pointer read FSessionID write SetSessionID;
    procedure FinishProc(ReferenceNumber: PMainChar; Status: Integer;
      StatusText: PMainChar; UPO: PMainChar; UserBreak: Boolean; Err: Integer;
      ErrMsg: PMainChar);
    procedure AddJPKXML(JPKXML:TCustomJPKXML);
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    property LicText:String read GetLicText write SetLicText;
    property ParentWinHandle:HWND read GetParentWinHandle write SetParentWinHandle;
    procedure Send(AWaitForEnd:Boolean=false); dynamic;
    procedure SendBreak;
    property TestMode:Boolean read FTestMode write SetTestMode;
    property SyncOnFinisch:Boolean read FSyncOnFinisch write SetSyncOnFinisch;
    property IsUserBreak:Boolean read GetIsUserBreak;
    property OnFinishSend:TJPKXMLFinishSend read FOnFinishSend write FOnFinishSend;
    function IsSending:Boolean;
  end;

  TJPKXMLSenderList = class
  private
    FList: TList;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(JPKXMLSender:TCustomJPKXMLSender);
    procedure Remove(JPKXMLSender:TCustomJPKXMLSender);
    function FindBySessionID(ASessionID:Pointer; out JPKXMLSender:TCustomJPKXMLSender):Boolean;
    function Count:Integer;
  end;

  TJPKXMLOneFileSender = class(TCustomJPKXMLSender)
  private
    procedure SetJPKXML(const Value: TCustomJPKXML);
    function GetJPKXML: TCustomJPKXML;
  public
    procedure Send(WaitForEnd:Boolean=false); override;
  published
    property LicText;
    property TestMode;
    property SyncOnFinisch;
    property OnFinishSend;
    property JPKXML:TCustomJPKXML read GetJPKXML write SetJPKXML;
  end;

procedure Register;

implementation

{$I eJPK_Interface.inc}

var
  vJPKXMLSenderSendList: TJPKXMLSenderList;
  vLicText: MainString = '';
  vParentWinHandle: HWND = 0;
const
  JPKType: array[TDocumentType] of Byte = (0,1);

procedure Register;
begin
  RegisterComponents('JPKXML', [TJPKXMLFile,TJPKXMLOneFileSender]);
end;

procedure InitJPKDLL_;
begin
  {$IFDEF JPKDynLink}
  if vJPKLib=0 then
    InitJPKDLL;
  {$ENDIF}
  if vLicText<>'' then
    SetLic(PMainChar(vLicText));
end;

procedure FinishProcGlobal(SessionID: Pointer;
    ReferenceNumber: PMainChar; Status: Integer;
    StatusText: PMainChar; UPO: PMainChar; UserBreak: Boolean; Err: Integer;
    ErrMsg: PMainChar); stdcall;
var
  JPKXMLSender: TCustomJPKXMLSender;
begin
  if vJPKXMLSenderSendList.FindBySessionID(SessionID,JPKXMLSender) then
    JPKXMLSender.FinishProc(ReferenceNumber,Status,StatusText,UPO,UserBreak,Err,ErrMsg)
  else
    raise Exception.Create('Bd krytyczny - nie znaleziono obiektu JPKXMLSender.');
end;

{ TCustomJPKXML }

{$IFNDEF UNICODE}
function StrLenW(wc:PWideChar):Integer;
var
  f: Integer;
begin
  Result:=0;
  if wc<>nil then
    for f:=0 to 10000 do
    begin
      if wc[f]=#0 then
      begin
        Result:=f;
        Exit;
      end;
    end;
end;
{$ENDIF}

function MainStrPas(pc:PMainChar):String;
{$IFNDEF UNICODE}
var
  ws: WideString;
{$ENDIF}
begin
  {$IFDEF UNICODE}
  if pc<>nil then
    Result:=pc
  else
    Result:='';
  {$ELSE}
  if pc<>nil then
  begin
    SetLength(ws,StrLenW(pc));
    Move(pc^,ws[1],Length(ws)*2);
    Result:=String(ws);
  end else
    Result:='';
  {$ENDIF}
end;

procedure TCustomJPKXML.AddError(AErr: Integer; AErrMsg: String);
begin
  with Error do
  begin
    err:=AErr;
    errMsg:=AErrMsg;
  end
end;

procedure TCustomJPKXML.AddJPKError;
var
  vErrMsg: pMainChar;
begin
  With Error do
  begin
    err:=JPKGetLastError(vErrMsg);
    errMsg:=MainStrPas(vErrMsg);
  end;
end;

destructor TCustomJPKXML.Destroy;
begin
  FreeAndnil(FErr);
  inherited;
end;

function TCustomJPKXML.GetError: TError;
begin
  if FErr=nil then
    FErr:=TError.Create;
  Result:=FErr;
end;

function TCustomJPKXML.GetLicText: String;
begin
  Result:=vLicText
end;

function TCustomJPKXML.GetParentWinHandle: HWND;
begin
  Result:=vParentWinHandle
end;

procedure TCustomJPKXML.RefreshUPO(Break:Boolean);
var
  vReferenceNumber: MainString;
  vUPO,vStatusText: PMainChar;
  vStatus: Integer;
begin
  InitJPKDLL_;

  vReferenceNumber:=FReferenceNumber;
  if Break then
  begin
    BreakGetUPO(PMainChar(vReferenceNumber));
    Exit;
  end;

  if FLockRefreshUPO then
    Exit;
  FLockRefreshUPO:=true;
  try

    if LowerCase(Copy(vReferenceNumber,1,5))='test:' then
    begin
      TestMode:=true;
      vReferenceNumber:=Copy(vReferenceNumber,6,maxint);
    end;
    if GetUPO(TestMode,PMainChar(FReferenceNumber),vUPO,vStatus,vStatusText) then
    begin
      FUPO:=MainStrPas(vUPO);
      FStatus:=vStatus;
      FStatusText:=vStatusText;
    end else
    begin
      AddJPKError;
      raise Exception.Create(Error.errMsg);
    end;
  finally
    FLockRefreshUPO:=false;
  end;
end;

procedure TCustomJPKXML.SetDocumentType(const Value: TDocumentType);
begin
  FDocumentType := Value;
end;

procedure TCustomJPKXML.SetLicText(const Value: String);
begin
  vLicText:=Value
end;

procedure TCustomJPKXML.SetParentWinHandle(const Value: HWND);
begin
  vParentWinHandle:=Value
end;

procedure TCustomJPKXML.SetReferenceNumber(const Value: MainString);
begin
  FReferenceNumber := Value;
end;

procedure TCustomJPKXML.SetStatus(const Value: Integer);
begin
  FStatus := Value;
end;

procedure TCustomJPKXML.SetStatusText(const Value: String);
begin
  FStatusText := Value;
end;

procedure TCustomJPKXML.SetTestMode(const Value: Boolean);
begin
  FTestMode := Value;
end;

procedure TCustomJPKXML.SetUPO(const Value: String);
begin
  FUPO := Value;
end;

{ TCustomJPKXMLSender }

procedure TCustomJPKXMLSender.AddJPKXML(JPKXML: TCustomJPKXML);
begin
  Assert(JPKXML<>nil);
  FJPKXMLList.Add(JPKXML);
end;

constructor TCustomJPKXMLSender.Create(AOwner: TComponent);
begin
  inherited;
  FJPKXMLList:=TJPKXMLList.Create(true);  
end;

destructor TCustomJPKXMLSender.Destroy;
begin
  FJPKXMLList.Free;
  inherited;
end;

procedure TCustomJPKXMLSender.FinishProc(ReferenceNumber: PMainChar;
  Status: Integer; StatusText, UPO: PMainChar; UserBreak: Boolean; Err: Integer;
  ErrMsg: PMainChar);
begin
  Inc(FFinishIndex);
  if FFinishIndex>=FJPKXMLList.Count then
    raise Exception.Create('Bad krytyczny - Ilo FinishProc przekracza ilo plikw XML wysanych w jednej sesji!');
  FJPKXMLList[FFinishIndex].ReferenceNumber:=MainStrPas(ReferenceNumber);
  FJPKXMLList[FFinishIndex].Status:=Status;
  FJPKXMLList[FFinishIndex].StatusText:=MainStrPas(StatusText);
  FJPKXMLList[FFinishIndex].AddError(err,MainStrPas(ErrMsg));
  FIsUserBreak:=UserBreak;

  if Assigned(FOnFinishSend) then
    FOnFinishSend(Self,FJPKXMLList[FFinishIndex]);

  if FFinishIndex=FJPKXMLList.Count-1 then
  begin
    vJPKXMLSenderSendList.Remove(Self);
    FSessionID:=nil;
  end;
  if vJPKXMLSenderSendList.Count=0 then
    FreeAndNil(vJPKXMLSenderSendList);
end;

function TCustomJPKXMLSender.GetIsUserBreak: Boolean;
begin
  Result:=FIsUserBreak;
end;

function TCustomJPKXMLSender.GetLicText: String;
begin
  Result:=vLicText
end;

function TCustomJPKXMLSender.GetParentWinHandle: HWND;
begin
  Result:=vParentWinHandle
end;

function TCustomJPKXMLSender.IsSending: Boolean;
begin
  Result:=SessionID<>nil;
end;

procedure TCustomJPKXMLSender.Send(AWaitForEnd:Boolean);
var
  f: Integer;
  vErrMsg: PMainChar;
  vRefNums: PMainChar;
begin
  FIsUserBreak:=false;
  vErrMsg:=nil;
  try

  InitJPKDLL_;

  if vJPKXMLSenderSendList<>nil then
    vJPKXMLSenderSendList.Remove(Self);

  SetParentwindowHandle(ParentWinHandle);
  FFinishIndex:=-1;
  for f := 0 to FJPKXMLList.Count-1 do
    With FJPKXMLList[f] do
    if IsIntData then
    begin
      if AddXMLJPK(PMainAnsiChar(GetData_),JPKType[DocumentType],PMainChar(GetFileName_))<0 then
      begin
        JPKGetLastError(vErrMsg);
        Exit;
      end;
    end else
    begin
      if AddXMLJPKFile(PMainChar(GetFileName_),JPKType[DocumentType])<0 then
      begin
        JPKGetLastError(vErrMsg);
        Exit;
      end;
    end;

  if vJPKXMLSenderSendList=nil then
    vJPKXMLSenderSendList:=TJPKXMLSenderList.Create;
  vJPKXMLSenderSendList.Add(Self);
  SessionID:=SignSendJPK(TestMode,nil,FinishProcGlobal,SyncOnFinisch);
  if SessionID=nil then
  begin
    Case JPKGetLastError(vErrMsg) of
      -10: raise TJPKXMLNoXMLJPKException.Create('Brak dokumentw XML');
    End;

  end;
  if AWaitForEnd then
    WaitForEnd(SessionID,vRefNums);

  finally
    if MainStrPas(vErrMsg)<>'' then
      raise Exception.Create(MainStrPas(vErrMsg));
  end;
end;

procedure TCustomJPKXMLSender.SendBreak;
begin
  SignSendBreak(FSessionID);
end;

procedure TCustomJPKXMLSender.SetLicText(const Value: String);
begin
  vLicText:=Value;
end;

procedure TCustomJPKXMLSender.SetParentWinHandle(const Value: HWND);
begin
  vParentWinHandle:=Value
end;

procedure TCustomJPKXMLSender.SetSessionID(const Value: Pointer);
begin
  FSessionID := Value;
end;

procedure TCustomJPKXMLSender.SetSyncOnFinisch(const Value: Boolean);
begin
  FSyncOnFinisch := Value;
end;

procedure TCustomJPKXMLSender.SetTestMode(const Value: Boolean);
var
  f: Integer;
begin
  FTestMode := Value;
  for f := 0 to FJPKXMLList.Count-1 do
    FJPKXMLList[f].TestMode:=FTestMode;
end;

{ TJPKXMLFile }

function TJPKXMLFile.GetData_: MainAnsiString;
begin
  Result:='';
end;

function TJPKXMLFile.GetFileName_: MainString;
begin
  Result:=FFileNamePath;
end;

function TJPKXMLFile.IsIntData: Boolean;
begin
  Result:=false;
end;

procedure TJPKXMLFile.SetFileNamePath(const Value: MainString);
begin
  FFileNamePath := Value;
end;

{ TJPKXMLList }

procedure TJPKXMLList.Add(JPKXML: TCustomJPKXML);
begin
  FList.AddObject('',JPKXML);
end;

procedure TJPKXMLList.Clear;
var
  f: Integer;
begin
  if FOwnsObj then  
    for f := FList.Count-1 downto 0 do
      FList.Objects[f].Free;
  FList.Clear;
end;

function TJPKXMLList.Count: Integer;
begin
  Result:=FList.Count
end;

constructor TJPKXMLList.Create(AOwnsObj:Boolean);
begin
  inherited Create;
  FOwnsObj:=AOwnsObj;
  FList:=TStringList.Create;
end;

procedure TJPKXMLList.Delete(index: Integer);
var
  vJPKXML: TCustomJPKXML;
begin
  if FOwnsObj then
    vJPKXML:=Items[index]
  else
    vJPKXML:=nil;

  FList.Delete(index);

  if FOwnsObj then
    FreeAndNil(vJPKXML);
end;

destructor TJPKXMLList.Destroy;
begin
  Clear;
  FList.Free;
  inherited;
end;

function TJPKXMLList.GetItems(index: Integer): TCustomJPKXML;
begin
  Result:=TCustomJPKXML(FList.Objects[index]);
end;

{ TJPKXMLSenderList }

procedure TJPKXMLSenderList.Add(JPKXMLSender: TCustomJPKXMLSender);
begin
  FList.Add(JPKXMLSender)
end;

function TJPKXMLSenderList.Count: Integer;
begin
  Result:=FList.Count
end;

constructor TJPKXMLSenderList.Create;
begin
  inherited;
  FList:=TList.Create;
end;

destructor TJPKXMLSenderList.Destroy;
begin
  FList.Free;
  inherited;
end;

function TJPKXMLSenderList.FindBySessionID(ASessionID: Pointer;
  out JPKXMLSender: TCustomJPKXMLSender): Boolean;
var
  f: Integer;
begin
  Result:=false;
  JPKXMLSender:=nil;
  for f := 0 to FList.Count-1 do
    if TCustomJPKXMLSender(FList[f]).SessionID=ASessionID then
    begin
      Result:=true;
      JPKXMLSender:=TCustomJPKXMLSender(FList[f]);
      Exit;
    end;      
end;

procedure TJPKXMLSenderList.Remove(JPKXMLSender: TCustomJPKXMLSender);
begin
  FList.Remove(JPKXMLSender)
end;

{ TJPKXMLOneFileSender }

function TJPKXMLOneFileSender.GetJPKXML: TCustomJPKXML;
begin
  if FJPKXMLList.Count>0 then
    Result:=FJPKXMLList[0]
  else
    Result:=nil;
end;

procedure TJPKXMLOneFileSender.Send(WaitForEnd:Boolean);
begin
  try
    inherited;
  except
    on E:TJPKXMLNoXMLJPKException do
    begin
      if JPKXML=nil then
        raise TJPKXMLNoXMLJPKException.Create('Brak property JPKXML!')
      else
      if JPKXML.IsIntData then
      begin
        if JPKXML.GetData_='' then
          raise TJPKXMLNoXMLJPKException.Create('Brak XML!')
      end else
        if JPKXML.GetFileName_='' then
          raise TJPKXMLNoXMLJPKException.Create('Nie wprowadzono nazwy pliku XML!')
    end else
      raise;
  end;
end;

procedure TJPKXMLOneFileSender.SetJPKXML(const Value: TCustomJPKXML);
begin
  if FJPKXMLList.Count>0 then
  begin
    FJPKXMLList.FOwnsObj:=false;
    FJPKXMLList.Delete(0);
  end;
  AddJPKXML(Value);
end;

initialization
finalization
  FreeAndNil(vJPKXMLSenderSendList);
end.
